#include <Pilot.h>
#include <SysEvtMgr.h>
#include <ScrDriver.h>
#include <FeatureMgr.h>			//	Needed to detect ROM versions
#include "Graytest_res.h"


/***********************************************************************
 *   Entry Points
 ***********************************************************************/


/***********************************************************************
 *   Internal Structures
 ***********************************************************************/
typedef struct {
    FontID appFontID;
    Boolean saveBackup;
}AddrPreferenceType;

typedef struct {
    UInt					renamedCategories;	// bitfield of categories with a different name
    char 					categoryLabels[dmRecNumCategories][dmCategoryLength];
    Byte 					categoryUniqIDs[dmRecNumCategories];
    Byte					lastUniqID;	// Uniq IDs generated by the device are between
    // 0 - 127.  Those from the PC are 128 - 255.
    Byte					reserved1;	// from the compiler word aligning things
    Word					reserved2;
    CountryType 		country;		// Country the database (labels) is formatted for
} AddrAppInfoType;

typedef AddrAppInfoType* AddrAppInfoPtr;


/***********************************************************************
 *   Global variables
 ***********************************************************************/
static Boolean HideSecretRecords;
static DmOpenRef AddrDB;
static FontID AppFontID = stdFont;
static Boolean SaveBackup = true;


/***********************************************************************
 *   Internal Constants
 ***********************************************************************/
//TODO Change the file creator ID'
#define appFileCreator             'APPL'
#define appVersionNum              0x01
#define appPrefID                  0x00
#define appPrefVersionNum          0x01
#define appDBName                  "AddrDB"
#define appDBType                  'DATA'
// Version Checking define
#define version20                  0x02000000


/***********************************************************************
 *   Internal Functions
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  formId - id of the form to display
 *
 * RETURNED:    VoidPtr
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static VoidPtr GetObjectPtr(Word objectID )
{
    FormPtr frmP;


    frmP = FrmGetActiveForm();
    return (FrmGetObjectPtr( frmP, FrmGetObjectIndex( frmP, objectID)));
}


/***********************************************************************
 *
 * FUNCTION: AddrAppInfoGetPtr
 *
 * DESCRIPTION: Return a locked pointer to the AppInfo or NULL
 *
 * PARAMETERS: dbP - open database pointer
 *
 * RETURNS: locked ptr to the AppInfo or NULL
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
AddrAppInfoPtr AddrAppInfoGetPtr(DmOpenRef *dbP )
{
    UInt cardNo;
    LocalID dbIDID;
    LocalID appInfoID;


    if (DmOpenDatabaseInfo(dbP, &dbIDID, NULL, NULL, &cardNo, NULL))
        return NULL;
    if (DmDatabaseInfo( cardNo, dbIDID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
        return NULL;
    if (appInfoID == NULL)
        return NULL;
    else
        return MemLocalIDToLockedPtr( appInfoID, cardNo);
}


/***********************************************************************
 *
 *  FUNCTION: AddrAppInfoInit
 *
 *  DESCRIPTION: Create an app info chunk if missing.  Set
 *		the strings to a default.
 *
 *  PARAMETERS: db - open database pointer
 *
 *  RETURNS: 0 if successful, errorcode if not
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
Err	AddrAppInfoInit(DmOpenRef *dbP )
{
    UInt cardNo;
    LocalID dbID;
    LocalID appInfoID;
    Handle tempH;
    AddrAppInfoPtr appInfoP;
    AddrAppInfoPtr defaultApplicationInfoP;
    Int loop;


    appInfoP = AddrAppInfoGetPtr(dbP);
    // If there isn't an ApplicationInfo make space for one
    if (appInfoP == NULL) {
        if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
            return dmErrInvalidParam;

        if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
            return dmErrInvalidParam;

        tempH = DmNewHandle(dbP, sizeof(AddrAppInfoType));
        if (!tempH) 
            return dmErrMemError;

        appInfoID = MemHandleToLocalID( tempH);
        DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);
        appInfoP = MemHandleLock(tempH);
    }


    // Allocate & Clear the app info
    defaultApplicationInfoP = MemPtrNew(sizeof(AddrAppInfoType));
    if (defaultApplicationInfoP == NULL) {
        ErrDisplay("Unable to init AddrDB");
        return 1;
    }


    MemSet( defaultApplicationInfoP, sizeof(AddrAppInfoType), 0);
    // Init the categories
    for (loop = 0; loop < dmRecNumCategories; loop++) {
        defaultApplicationInfoP->categoryUniqIDs[loop] = loop;
    }


    defaultApplicationInfoP->lastUniqID = dmRecNumCategories - 1;
    // copy in the defaults and free the default app info
    DmWrite(appInfoP, 0, defaultApplicationInfoP,  sizeof(AddrAppInfoType));
    MemPtrFree(defaultApplicationInfoP);
    // Try to use localized app info block strings.
    //AddrLocalizeAppInfo(appInfoP);
    // Localize the field labels to the current country
    //AddrChangeCountry(appInfoP);
    // Unlock
    MemPtrUnlock(appInfoP);
    return 0;
}


/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version meets your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags that indicate if the application 
 *                                UI is initialized.
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 ***********************************************************************/
static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags)
{
    DWord romVersion;


    // See if we have at least the minimum required version of the ROM or later.
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion) {
        if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
            // TODO: Display an Alert with an alert UI
            // FrmAlert(<RomIncompatibleAlert>);
            // Pilot 1.0 will continuously relaunch this app unless we switch to 
            // another safe one.
            if (romVersion < version20) {
               // Err err;


                AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
            }
        }
        return sysErrRomIncompatible;
    }
    return 0;
}
/***********************************************************************
 *
 * FUNCTION:    AppHandleSync
 *
 * DESCRIPTION: Handle details after the database has been synchronized.
 *
 * PARAMETERS:    findParams
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppHandleSync(void)
{
    DmOpenRef dbP;
//    AddrAppInfoPtr appInfoP;
    UInt mode;


    // Find the application's data file.
    mode = dmModeReadWrite;
    dbP = DmOpenDatabaseByTypeCreator(appDBType, sysFileCAddress, mode);
    if (!dbP) 
        return;

    DmCloseDatabase(dbP);
}


/***********************************************************************
 *
 * FUNCTION:    Search
 *
 * DESCRIPTION: This routine searchs the application's database for records
 *              contains the string passed. 
 *
 * PARAMETERS:    findParams
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void Search(FindParamsPtr paramsP)
{
   AddrAppInfoPtr appInfoP = 0;
//    AddrDBRecordType     record;
    Boolean done = false;
    Boolean match = false;
 //   CharPtr headerP;
    CharPtr unnamedRecordStringP = NULL;
    DmOpenRef dbP;
    DmSearchStateType    searchState;
    Err error = 0;
  //  Handle headerStringH;
    LocalID dbID;
    RectangleType rect;
    UInt cardNo = 0;
    UInt recordNum;
    VoidHand recordH;
    Word loop = 0;
   Word pos = 0;


    // Find the application's data file.
    error = DmGetNextDatabaseByTypeCreator(true, &searchState, appDBType    , sysFileCAddress, true, &cardNo, &dbID);
    if (error) {
        paramsP->more = false;
        return;
    }
    // Open the application's database.
    dbP = DmOpenDatabase(cardNo, dbID, paramsP->dbAccesMode);
    if (! dbP) {
        paramsP->more = false;
        return;
    }

    // Display the heading line.
    /*
    headerStringH = DmGetResource(strRsc, FindAddrHeaderStr);
    headerP = MemHandleLock(headerStringH);
    done = FindDrawHeader (paramsP, headerP);
    MemHandleUnlock(headerStringH);
    DmReleaseResource(headerStringH);   
    */
    if (done) 
        goto Exit;
    // Search the description and note fields for the "find" string.
    recordNum = paramsP->recordNum;
    while (true) {
        // Because applications can take a long time to finish a find when
        // the result may be on the screen or for other reasons, users like
        // to be able to stop the find.  Stop the find if an event is pending.
        // This stops if the user does something with the device.  Because
        // this call slows down the search we perform it every so many 
        // records instead of every record.  The response time should still
        // be short without introducing much extra work to the search.
        // Note that in the implementation below, if the next 16th record is  
        // secret the check doesn't happen.  Generally this shouldn't be a 
        // problem since if most of the records are secret then the search 
        // won't take long anyways!
        if ((recordNum & 0x000f) == 0 &&         // every 16th record
            EvtSysEventAvail(true)) {
            // Stop the search process.
            paramsP->more = true;
            break;
        }

        recordH = DmQueryNextInCategory(dbP, &recordNum, dmAllCategories);
        // Have we run out of records?
        if (! recordH) {
            paramsP->more = false;
            break;
        }
        // Search all the fields of the Addr's record.
        /*
        AddrGetRecord(dbP, recordNum, &record, &recordH);
        for (loop = 0; loop < AddrNumFields; loop++) {
            if (record.fields[loop]) {
                match = FindStrInStr(record.fields[loop], paramsP->strToFind, &pos);
                if (match) 
                    break;

            }


        }


        */
        if (match) {
            done = FindSaveMatch(paramsP, recordNum, pos, loop, 0, cardNo, dbID);
            if (done) {
                MemHandleUnlock(recordH);
                break;

            }


            // Get the bounds of the region where we will draw the results.
            FindGetLineBounds(paramsP, &rect);
            //appInfoP = (AddrAppInfoPtr) AddrAppInfoGetPtr(dbP);
            // TODO: Display the title of the description.
            MemPtrUnlock(appInfoP);
            paramsP->lineNumber++;
        }


        MemHandleUnlock(recordH);
        recordNum++;
    }


    if (unnamedRecordStringP) {
        MemPtrUnlock(unnamedRecordStringP);
        //DmReleaseResource(strRsc, unnamedRecordStringP)
    }


Exit:
    DmCloseDatabase(dbP);
}


/***********************************************************************
 *
 * FUNCTION:    AppGoToItem
 *
 * DESCRIPTION: This routine is a entry point of this application.
 *              It is generally call as the result of hiting of 
 *              "Go to" button in the text search dialog.
 *
 * PARAMETERS:    recordNum - 
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppGoToItem(GoToParamsPtr paramsP, Boolean launchingApp)
{
    Word formID = 0;
    Word recordNum;
    UInt attr;
    ULong uniqueID;
    EventType event;


    recordNum = paramsP->recordNum;
    DmRecordInfo(AddrDB, recordNum, &attr, &uniqueID, NULL);


    // Change the current category if necessary.
    //if (<App>CurrentCategory != dmAllCategories) {
        //<App>CurrentCategory = attr & dmRecAttrCategoryMask;
    //}
    // If the application is already running, close all the open forms.  If
    // the current record is blank, then it will be deleted, so we'll 
    // the record's unique id to find the record index again, after all 
    // the forms are closed.
    if (!launchingApp) {
        FrmCloseAllForms();
        DmFindRecordByID(AddrDB, uniqueID, &recordNum);
    }


    // Set global variables that keep track of the currently record.


    //<App>CurrentRecord = recordNum;
    // TODO: Determine the view for the record
    // formID = ...
    MemSet(&event, sizeof(EventType), 0);

    // Send an event to load the form we want to goto.
    event.eType = frmLoadEvent;
    event.data.frmLoad.formID = formID;
    //EvtAddEventToQueue(&event);

    // Send an event to goto a form and select the matching text.
    event.eType = frmGotoEvent;
    event.data.frmGoto.formID = formID;
    event.data.frmGoto.recordNum = recordNum;
    event.data.frmGoto.matchPos = paramsP->matchPos;
    event.data.frmGoto.matchLen = paramsP->searchStrLen;
    event.data.frmGoto.matchFieldNum = paramsP->matchFieldNum;
    // EvtAddEventToQueue (&event);
}


/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:  frm - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void MainFormInit(FormPtr frmP)
{
    unsigned long scrDepth = 2;
	ScrDisplayMode(scrDisplayModeSet, 0, 0, &scrDepth, 0);
	FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, scrDepth);
}

/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean MainFormDoCommand(Word command)
{
  //  FieldPtr fldP;
    Boolean handled = true;


    switch (command) {
        default:
            handled = false;
            break;

    }


    return handled;
}




/***********************************************************************
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the
 *              "MainForm" of this application.
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent( EventPtr eventP)
{
    Boolean handled = false;
    FormPtr frmP;

    switch (eventP->eType) {
        case ctlSelectEvent:
            switch (eventP->data.ctlSelect.controlID) {
                case RomIncompatibleOK:
                    break;

                default:
                    break;
            }
            break;

        case menuEvent: 
            return MainFormDoCommand(eventP->data.menu.itemID);

        case keyDownEvent:

            switch (eventP->data.keyDown.chr) {
                case pageUpChr:
                    break;
                case pageDownChr:
                    break;
                case linefeedChr:
                    break;
                default:
                    break;
            }
            break;
        case frmOpenEvent:
            frmP = FrmGetActiveForm();
            MainFormInit( frmP);
            FrmDrawForm ( frmP);
            handled = true;
            break;

        case frmUpdateEvent:
            /*
            MainFormUpdateDisplay( eventP->data.frmUpdate.updateCode);
            handled = true;
            */
            break;

        default:
            break;
    }

    return handled;
}

/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean AppHandleEvent( EventPtr eventP)
{
    Word formId;
    FormPtr frmP;


    if (eventP->eType == frmLoadEvent) {
        // Load the form resource.
        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm( formId);
        FrmSetActiveForm( frmP);
        // Set the event handler for the form.  The handler of the currently
        // active form is called by FrmHandleEvent each time is receives an
        // event.
        switch (formId) { //{{PFW_FORM_LOAD
            case MainForm:
                FrmSetEventHandler(frmP, MainFormHandleEvent);
                break;
            default:
                //ErrNonFatalDisplay("Invalid Form Load Event");
                break;

        } //}}PFW_FORM_LOAD

        return true;
    }

    return false;
}

/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppEventLoop(void)
{
    Word error;
    EventType event;


    do {
        EvtGetEvent(&event, evtWaitForever);
        if (! SysHandleEvent(&event))
            if (! MenuHandleEvent(0, &event, &error))
                if (! AppHandleEvent(&event))
                    FrmDispatchEvent(&event);
    } while (event.eType != appStopEvent);
}

/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  This routine opens the application's resource file and
 *               database.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value 0 if nothing went wrong
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Err AppStart(void)
{
    Err error= 0;
    UInt mode;
    SystemPreferencesType sysPrefs;
    AddrPreferenceType prefs;
    Word prefsSize;
    AddrAppInfoPtr appInfoP;


    // Determime if secert record should be shown.
    PrefGetPreferences(&sysPrefs);
    HideSecretRecords = sysPrefs.hideSecretRecords;
    if (sysPrefs.hideSecretRecords)
        mode = dmModeReadWrite;
    else
        mode = dmModeReadWrite | dmModeShowSecret;
    // Find the application's data file.  If it doesn't exist create it.
    AddrDB = DmOpenDatabaseByTypeCreator( appDBType, sysFileCAddress, mode);
    if (!AddrDB) {
        error = DmCreateDatabase(0, appDBName, sysFileCAddress, appDBType, false);
        if (error) 
            return error;

        AddrDB = DmOpenDatabaseByTypeCreator(appDBType, sysFileCAddress, mode);
        if (!AddrDB)
            return 1;

        error = AddrAppInfoInit(AddrDB);
        if (error) 
            return error;
    }

    appInfoP = (AddrAppInfoPtr) AddrAppInfoGetPtr(AddrDB);
    ErrFatalDisplayIf(appInfoP == NULL, "Missing app info block");
    // Update the database to look and behave properly for the given country.
    if (appInfoP->country != sysPrefs.country) {
        // AddrChangeCountry(appInfoP);
        // Read the preferences / saved-state information.  There is only one
        // version of the Applications preferences so don't worry about multiple
        // versions.
        prefsSize = sizeof(AddrPreferenceType);
        if (PrefGetAppPreferences(sysFileCAddress, appPrefID, &prefs, &prefsSize, true) != 
                noPreferenceFound) {
            AppFontID = prefs.appFontID;
            SaveBackup = prefs.saveBackup;
        }

    }

    MemPtrUnlock(appInfoP);
    return error;
}

/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: This routine close the application's database
 *              and save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppStop(void)
{
}

/***********************************************************************
 * FUNCTION:    AddrPilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static DWord AddrPilotMain( Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Err error;


    error = RomVersionCompatible(version20, launchFlags);
    if (error) 
        return error;

    switch (cmd) {
        case sysAppLaunchCmdNormalLaunch:
            error = AppStart();
            if (error) 
                return error;
            FrmGotoForm(MainForm);
            AppEventLoop();
            AppStop();
            break;
        case sysAppLaunchCmdFind:
            Search((FindParamsPtr) cmdPBP);
            break;
        case sysAppLaunchCmdGoTo:
            {
                Boolean launched;


                launched = launchFlags & sysAppLaunchFlagNewGlobals;
                if (launched) {
                    error = AppStart();
                    if (error) 
                        return error;
                }
                AppGoToItem((GoToParamsPtr)cmdPBP, launched);
                if (launched) {
                    AppEventLoop();
                    AppStop();   
                }
            }
            break;
        case sysAppLaunchCmdSyncNotify:
            AppHandleSync();
            break;
        case sysAppLaunchCmdSaveData:
            FrmSaveAllForms();
            break;
        case sysAppLaunchCmdInitDatabase:
            //AppLaunchCmdDatabaseInit(((SysAppLaunchCmdInitDatabaseType*)cmdPBP)->dbP);
            break;
        case sysAppLaunchCmdSystemReset:
            if (((SysAppLaunchCmdSystemResetType*)cmdPBP)->createDefaultDB) {
                Handle resH;


                resH = DmGet1Resource(sysResTDefaultDB, sysResIDDefaultDB);
                if (resH) {
                    DmCreateDatabaseFromImage(MemHandleLock(resH));
                    MemHandleUnlock(resH);
                    DmReleaseResource(resH);
                }
            }
            break;
        default:
            break;
    }
    return 0;
}
/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
DWord PilotMain( Word cmd, Ptr cmdPBP, Word launchFlags)
{
    return AddrPilotMain( cmd, cmdPBP, launchFlags);
}

